import logging
import os
import random
import re
from multiprocessing import Pool
from typing import Optional, List

from time import sleep

from batched_sql_context import BatchedSqlContext
from deprov_config import config
from service import Services
from utils import timed_sql_operation


class AsyncBatchConfig(object):
    def __init__(self,
                 context: BatchedSqlContext,
                 name: str,
                 batch_size: int,
                 batched_table: str,
                 batch_query: str,
                 pool_size: Optional[int] = None,
                 dry_run: bool = False,
                 timed: bool = False):
        self.context = context
        self.name = name
        self.batch_size = batch_size
        self.batched_table = batched_table
        self.batch_query = batch_query
        self.pool_size = pool_size
        self.dry_run = dry_run
        self.timed = timed


class ForeignKeyViolation(object):
    expression = re.compile(r'.*update or delete on table "([^"]*)" violates .*\n.*Key \(([^)]*)\)=\(([^)]*)')

    def __init__(self, ex: Exception):
        match = self.expression.match(str(ex))
        (self.table, self.field, self.value) = match.groups()

    def __str__(self):
        return f' AND {self.table}.{self.field} <> {self.value}'


class ForeignKeyRemoval(object):
    def __init__(self):
        self.violations: List[ForeignKeyViolation] = []

    def append(self, ex: Exception):
        self.violations.append(ForeignKeyViolation(ex))

    def __str__(self):
        if not self.violations:
            return ''
        return ''.join([str(violation) for violation in self.violations])


class AsyncBatchRunner(object):
    def __init__(self, runner_id: int, start: int, amount: int, batched_table: str, batch_query: str, dry_run: bool):
        self._id = runner_id
        self._start = start
        self._amount = amount
        self._batched_table = batched_table
        self._batch_query = batch_query
        self._dry_run = dry_run

    def run(self) -> int:
        config.vlog(f'>> ({self._id}) running...')
        config.vlog(f'''
Id:            {self._id}
Start:         {self._start:,}
Amount:        {self._amount:,}
Batched Table: {self._batched_table}
Batch Query:   {self._batch_query}
Dry Run:       {self._dry_run}
            ''')

        foreign_key_removal = ForeignKeyRemoval()
        run_count = 0
        while True:
            run_count += 1
            retry_wait = random.random() * run_count
            try:
                result = self._execute_sql(foreign_key_removal)
                if run_count > 1:
                    logging.warning(f'++ (PID: {os.getpid()} ID: {self._id} Attempt: {run_count}) Completed retried batch!')
                return result
            except Exception as ex:
                if 'DeadlockDetected' in type(ex).__name__:
                    logging.warning(f'(PID: {os.getpid()} ID: {self._id} Attempt: {run_count}) Deadlock Detected. Retrying in {retry_wait:.3f} seconds...')
                elif 'ForeignKeyViolation' in type(ex).__name__:
                    foreign_key_removal.append(ex)
                    logging.warning(f'(PID: {os.getpid()} ID: {self._id} Attempt: {run_count}) Foreign key violation detected. Retrying...')
                    retry_wait = 0
                else:
                    raise
            finally:
                async_db_connection.rollback()
            sleep(retry_wait)

    def _execute_sql(self, foreign_key_removal: ForeignKeyRemoval):
        with async_db_connection.cursor() as cursor:
            # We disable hashjoin which is slow as beans.
            cursor.execute(f'''
                SET LOCAL enable_hashjoin = off;
                SET LOCAL enable_mergejoin = off;
                DROP TABLE IF EXISTS batch_table;
                CREATE TEMP TABLE batch_table (LIKE {self._batched_table});
                INSERT INTO batch_table SELECT * FROM {self._batched_table} OFFSET {self._start} LIMIT {self._amount};
                {self._batch_query}{foreign_key_removal};''')
            row_count = cursor.rowcount
            cursor.execute('DROP TABLE batch_table;')
            if self._dry_run:
                async_db_connection.rollback()
            else:
                async_db_connection.commit()
        config.vlog(f'>> ({self._id}) complete [{row_count:,} rows].')
        return row_count


async_db_connection = None


def init_worker():
    if not config.verbose:
        logging.getLogger().handlers[0].setLevel(logging.WARNING)
    logging.info(f'Initializing worker ({os.getpid()}).')
    global async_db_connection
    async_db_connection = Services().open().db_connection


class AsyncBatchJob(object):
    def __init__(self, batch_config: AsyncBatchConfig):
        self._config = batch_config
        self._row_count = 0
        self._db_connection = None

    def _get_count(self):
        self._config.context.cursor.execute(f"SELECT count(*) FROM {self._config.batched_table};")
        count = self._config.context.cursor.fetchone()[0] or 0
        if not count:
            logging.info(f'{self._config.name}: Nothing to loop on.')
        return count

    def run(self):
        if self._config.timed:
            return self._run_timed()
        else:
            return self._run()

    @timed_sql_operation
    def _run_timed(self):
        return self._run()

    def _run(self):
        count = self._get_count()
        if not count:
            return 0

        runners = []
        runner_id = 0
        for start in range(0, count+1, self._config.batch_size):
            runner_id += 1
            runners.append(AsyncBatchRunner(runner_id, start, self._config.batch_size, self._config.batched_table, self._config.batch_query, self._config.dry_run))

        pool_size = self._config.pool_size or config.pool_size
        logging.info(F'Async Batching: {self._config.name} ({len(runners):,} runners and {pool_size:,} workers.)')
        result = Pool(processes=pool_size, initializer=init_worker).map(self._start_runner, runners)
        logging.info(F'Async Batching complete: {self._config.name}')
        return sum(result)

    @staticmethod
    def _start_runner(runner: AsyncBatchRunner):
        return runner.run()
