import logging
import os
import random
from multiprocessing import Pool
from typing import Optional, List, Tuple

import boto3

from batched_sql_context import BatchedSqlContext
from deprov_config import config
from service import Services
from aws.s3_client import S3Client, S3DeleteObjectsResult
from utils import timed_sql_operation


_DRY_RUN_ERROR_RATE = 0.10


class S3PurgeBatchConfig(object):
    def __init__(self,
                 context: BatchedSqlContext,
                 batch_size: int,
                 bucket_name: str,
                 pool_size: Optional[int] = None,
                 dry_run: bool = False,
                 timed: bool = False):
        self.context = context
        self.name = 'Async S3 Purge'
        self.bucket_name = bucket_name
        self.batch_size = batch_size
        self.pool_size = pool_size
        self.dry_run = dry_run
        self.timed = timed


class S3DryRun(object):
    def delete_files(self, bucket_name, key_names) -> S3DeleteObjectsResult:
        deleted = []
        errors = []
        for key in key_names:
            rand = random.random()
            if rand <= _DRY_RUN_ERROR_RATE:
                errors.append({'Key': key, 'Code': f'CODE{rand}', 'Message': f'MESSAGE{rand}'})
            else:
                deleted.append({'Key': key})
        return S3DeleteObjectsResult({'Deleted': deleted, 'Errors': errors})


class AsyncS3PurgeBatchRunner(object):
    def __init__(self, runner_id: int, start: int, amount: int, bucket_name: str, dry_run: bool):
        self._id = runner_id
        self._start = start
        self._amount = amount
        self._bucket_name = bucket_name
        self._dry_run = dry_run

    @timed_sql_operation
    def run(self) -> int:
        config.vlog(f'>> ({self._id}) running...')
        config.vlog(f'''
Id:            {self._id}
Start:         {self._start:,}
Amount:        {self._amount:,}
Dry Run:       {self._dry_run}
            ''')

        files_to_delete = self._get_file_info()
        result = self._delete_files(files_to_delete)
        self._clear_files(result, files_to_delete)
        return len(result.deleted_objects)

    def _get_file_info(self) -> List[Tuple[int, str]]:
        with async_db_connection.cursor() as cursor:
            cursor.execute(f'SELECT id, s3_key FROM file_details_to_delete_batch ORDER BY id OFFSET {self._start} LIMIT {self._amount};''')
            return [(row[0], row[1]) for row in cursor.fetchall()]

    def _delete_files(self, files_to_delete: List[Tuple[int, str]]) -> S3DeleteObjectsResult:
        if not files_to_delete:
            return S3DeleteObjectsResult()
        if self._dry_run:
            return S3DryRun().delete_files(self._bucket_name, [entry[1] for entry in files_to_delete])
        return s3_session.delete_files(self._bucket_name, [entry[1] for entry in files_to_delete])

    def _clear_files(self, deletion_result: S3DeleteObjectsResult, files_to_delete: List[Tuple[int, str]]):
        if deletion_result.errors:
            logging.warning(f'There were {len(deletion_result.errors)} errors when deleting {len(files_to_delete)} files.')

        exclusion_string = "".join([f" AND s3_key <> '{entry.key}'" for entry in deletion_result.errors])
        with async_db_connection.cursor() as cursor:
            cursor.execute(f'''
                UPDATE file_details_to_delete SET deleted_on = now() FROM (
                    SELECT id FROM file_details_to_delete_batch ORDER BY id OFFSET {self._start} LIMIT {self._amount}) AS current_batch
                    WHERE current_batch.id = file_details_to_delete.id{exclusion_string};''')
            for entry in deletion_result.errors:
                cursor.execute(f"UPDATE file_details_to_delete SET error_code = %s, error_message = %s WHERE s3_key = %s;", (entry.code, entry.message, entry.key))
            async_db_connection.commit()


async_db_connection = None
s3_session: Optional[S3Client] = None


def init_worker():
    if not config.verbose:
        logging.getLogger().handlers[0].setLevel(logging.WARNING)
    logging.info(f'Initializing worker ({os.getpid()}).')
    global async_db_connection, s3_session
    services = Services().open()
    async_db_connection = services.db_connection
    s3_session = services.s3_client()


class AsyncS3PurgeBatchJob(object):
    def __init__(self, batch_config: S3PurgeBatchConfig):
        self._config = batch_config
        self._row_count = 0
        self._db_connection = None

    def _get_count(self):
        self._config.context.cursor.execute(f"SELECT count(*) FROM file_details_to_delete_batch;")
        count = self._config.context.cursor.fetchone()[0] or 0
        if not count:
            logging.info(f'{self._config.name}: Nothing to loop on.')
        return count

    def run(self):
        if self._config.timed:
            return self._run_timed()
        else:
            return self._run()

    @timed_sql_operation
    def _run_timed(self):
        return self._run()

    def _run(self):
        count = self._get_count()
        if not count:
            return 0

        runners = []
        runner_id = 0
        for start in range(0, count+1, self._config.batch_size):
            runner_id += 1
            runners.append(AsyncS3PurgeBatchRunner(runner_id, start, self._config.batch_size, self._config.bucket_name, self._config.dry_run))

        pool_size = self._config.pool_size or config.pool_size
        logging.info(F'Async S3 Purge Batching: ({len(runners):,} runners and {pool_size:,} workers.)')
        result = Pool(processes=pool_size, initializer=init_worker).map(self._start_runner, runners)
        logging.info(F'Async S3 Purge Batching complete.')
        return sum(result)

    @staticmethod
    def _start_runner(runner: AsyncS3PurgeBatchRunner):
        return runner.run()
