import datetime
import functools
from typing import Optional

import logging

from service import Services
from batched_sql_context import BatchedSqlContext
from batch_logging import setup_logging
from batch_s3_purge import AsyncS3PurgeBatchJob, S3PurgeBatchConfig
from utils import timed_sql_operation
from deprov_config import config
from deprovision_base import Deprovision


class NothingToDoError(Exception):
    pass


class S3CleanupRunner(object):
    def __init__(self, batched_context: BatchedSqlContext):
        self._context = batched_context

    def run(self):
        if self._context.dry_run:
            logging.info('Performing dry run (will not make changes).')
        self._setup_db()
        self._drop_tables()
        self._create_tables()
        self._fill_batch_table()
        self._run_batch()

    @timed_sql_operation
    def _setup_db(self):
        logging.info('Setting up the db for deprovisioning.')
        self._context.cursor.execute('''
            CREATE TABLE IF NOT EXISTS orgs_to_deprovision(org_key TEXT PRIMARY KEY, priority INT DEFAULT 10, deleted_on TIMESTAMP DEFAULT NULL);
            CREATE INDEX IF NOT EXISTS orgs_to_deprovision_deleted_on_index ON orgs_to_deprovision (deleted_on);
            CREATE INDEX IF NOT EXISTS orgs_to_deprovision_priority_index ON orgs_to_deprovision (priority);
            -- This table will save the files that we need to delete from S3.
            CREATE TABLE IF NOT EXISTS file_details_to_delete(id BIGINT PRIMARY KEY, s3_key VARCHAR(1024) NOT NULL, file_index_id BIGINT, deleted_on TIMESTAMP DEFAULT NULL, error_code TEXT DEFAULT NULL, error_message TEXT DEFAULT NULL);
            CREATE INDEX IF NOT EXISTS file_details_to_delete_file_index_id_index ON file_details_to_delete (file_index_id);
            CREATE INDEX IF NOT EXISTS file_details_to_delete_deleted_on_index ON file_details_to_delete (deleted_on);
            CREATE INDEX IF NOT EXISTS file_details_to_delete_s3_key_index ON file_details_to_delete (s3_key);
            ''')

    @timed_sql_operation
    def _drop_tables(self):
        logging.info('Dropping tables.')
        self._context.cursor.execute('DROP TABLE IF EXISTS file_details_to_delete_batch;')

    @timed_sql_operation
    def _create_tables(self):
        logging.info('Creating tables.')
        self._context.cursor.execute('CREATE TABLE file_details_to_delete_batch(id BIGINT PRIMARY KEY, s3_key VARCHAR(1024) NOT NULL);')

    @timed_sql_operation
    def _fill_batch_table(self):
        logging.info('Filling batch table.')
        self._context.cursor.execute('INSERT INTO file_details_to_delete_batch SELECT id, s3_key FROM file_details_to_delete WHERE deleted_on IS NULL;')
        return self._context.cursor.rowcount

    def _run_batch(self):
        AsyncS3PurgeBatchJob(S3PurgeBatchConfig(
            context=self._context,
            batch_size=min(config.batch_size, 1000),
            bucket_name=config.s3_bucket_name,
            pool_size=config.pool_size,
            dry_run=config.dry_run,
            timed=True
        )).run()


def run():
    try:
        with Services(auto_commit=True) as services:
            logging.info('Starting S3 cleanup.')
            Deprovision(services.db_connection, config, S3CleanupRunner).run()
            logging.info('S3 cleanup completed successfully.')
    except NothingToDoError as e:
        logging.info(f'{e}  Exiting...')


def main():
    setup_logging(config)
    config.log()
    run()


if __name__ == "__main__":
    try:
        main()
    except Exception as ex:
        logging.exception(f'Runtime Error ({type(ex)}): {ex}')
        exit(1)
