import datetime
import functools
from typing import Optional

import logging

from service import Services
from async_batch import AsyncBatchConfig, AsyncBatchJob
from batched_sql_context import BatchedSqlContext
from batch_logging import setup_logging
from staged_run import StagedRun, DbStageStorageManager
from utils import timed_sql_operation
from deprov_config import config
from deprovision_base import Deprovision
from time import sleep


class NothingToDoError(Exception):
    pass


class DeprovisionRunner(object):
    def __init__(self, batched_context: BatchedSqlContext):
        self._context = batched_context
        self._org_keys = None
        self._staged_run = self._init_staged_run()

    def _init_staged_run(self):
        staged_run = StagedRun(DbStageStorageManager(self._context.cursor))
        with staged_run.add_stage('Initialize run') as stage:
            stage.append(self._setup_db)
            stage.append(self._clean_org_keys)
            stage.append(self._get_org_keys)
            stage.append(self._drop_tables)
            stage.append(self._create_tables)
        staged_run.add_stage('Find customer links to delete').append(self._find_cust_links_to_delete)
        staged_run.add_stage('Find customer devices').append(self._find_cust_devices)
        staged_run.add_stage('Find unique files to be deleted').append(self._find_unique_files_to_delete)
    #   No longer used since the UBS metadata / binary data split
        staged_run.add_stage('Find file details to be deleted').append(self._find_s3_files_to_delete)
        staged_run.add_stage('Copy unique hashes for later deletion').append(self._copy_file_hashes_to_delete)
        staged_run.add_stage('Delete customer devices').append(self._delete_devices)
        staged_run.add_stage('Delete file details').append(self._delete_file_details)
        staged_run.add_stage('Delete customer file index links').append(self._delete_cust_links)
        staged_run.add_stage('Delete metadata and file indexes').append(self._delete_metadata)
        staged_run.add_stage('Delete customers').append(self._delete_customers)
    #   No longer used since the UBS metadata / binary data split
        staged_run.add_stage('Copy file details for later deletion').append(self._copy_file_details_to_delete)
        staged_run.add_stage('Mark orgs as deleted').append(self._remove_org_keys_from_deprov_list)
        staged_run.add_stage('Cleanup db').append(self._drop_tables)
        return staged_run

    def run(self):
        if self._context.dry_run:
            logging.info('Performing dry run (will not be committed).')
        self._init_async()
        self._staged_run.run()

    def _init_async(self):
        logging.info('Initializing async run.')
        self._context.cursor.execute('CREATE TABLE IF NOT EXISTS async_orgs(org_key TEXT PRIMARY KEY);')
        if not self._staged_run.in_progress():
            self._context.cursor.execute('DELETE FROM async_orgs;')
        else:
            self._context.cursor.execute('SELECT org_key FROM async_orgs;')
            self._org_keys = [record[0] for record in self._context.cursor.fetchall()]
            if not self._org_keys:
                raise(Exception('Missing org keys from in progress async run.'))
            logging.info(f'The following org_keys were found to deprovision: {self._org_keys} (Count: {len(self._org_keys)})')

    def _setup_db(self):
        self._run_sql_query(
            'Setting up the db for deprovisioning.',
            '''
            ALTER TABLE cust_device_link DROP CONSTRAINT IF EXISTS cust_device_link_device_id_fkey;
            ALTER TABLE cust_device_link_1 DROP CONSTRAINT IF EXISTS cust_device_link_1_device_id_fkey;
            ALTER TABLE cust_device_link_3 DROP CONSTRAINT IF EXISTS cust_device_link_3_device_id_fkey;
            
            ALTER TABLE cust_device_link DROP CONSTRAINT IF EXISTS cust_device_link_cust_ix_link_id_fkey;
            ALTER TABLE cust_device_link_1 DROP CONSTRAINT IF EXISTS cust_device_link_1_cust_ix_link_id_fkey;
            ALTER TABLE cust_device_link_3 DROP CONSTRAINT IF EXISTS cust_device_link_3_cust_ix_link_id_fkey;
            
            ALTER TABLE file_details DROP CONSTRAINT IF EXISTS file_details_file_index_id_fkey;
            ALTER TABLE signature_result_link DROP CONSTRAINT IF EXISTS signature_result_link_signature_id_fkey;
            ALTER TABLE cust_filepath DROP CONSTRAINT IF EXISTS cust_filepath_filepath_id_fkey;
            
            ALTER TABLE cust_filepath DROP CONSTRAINT IF EXISTS cust_filepath_cust_ix_link_id_fkey;
            ALTER TABLE signature_result_link DROP CONSTRAINT IF EXISTS signature_result_link_cust_ix_link_id_fkey;
            
            ALTER TABLE file_details_magic_link DROP CONSTRAINT IF EXISTS file_details_magic_link_file_details_id_fkey;
            ALTER TABLE file_details_partner_link DROP CONSTRAINT IF EXISTS file_details_partner_link_file_details_id_fkey;

            CREATE TABLE IF NOT EXISTS orgs_to_deprovision(org_key TEXT PRIMARY KEY, priority INT DEFAULT 10, deleted_on TIMESTAMP DEFAULT NULL);
            CREATE INDEX IF NOT EXISTS orgs_to_deprovision_deleted_on_index ON orgs_to_deprovision (deleted_on);
            CREATE INDEX IF NOT EXISTS orgs_to_deprovision_priority_index ON orgs_to_deprovision (priority);
            -- This table was used to save the files that we need to delete from S3.
            CREATE TABLE IF NOT EXISTS file_details_to_delete(id BIGINT PRIMARY KEY, s3_key VARCHAR(1024) NOT NULL, file_index_id BIGINT, deleted_on TIMESTAMP DEFAULT NULL, error_code TEXT DEFAULT NULL, error_message TEXT DEFAULT NULL);
            CREATE INDEX IF NOT EXISTS file_details_to_delete_file_index_id_index ON file_details_to_delete (file_index_id);
            CREATE INDEX IF NOT EXISTS file_details_to_delete_deleted_on_index ON file_details_to_delete (deleted_on);
            CREATE INDEX IF NOT EXISTS file_details_to_delete_s3_key_index ON file_details_to_delete (s3_key);
            -- This table will save the hashes that we need to delete from S3.
            CREATE TABLE IF NOT EXISTS file_hashes_to_delete(id BIGINT PRIMARY KEY, sha256 VARCHAR(256) NOT NULL, deleted_on TIMESTAMP DEFAULT NULL);
            CREATE INDEX IF NOT EXISTS file_hashes_to_delete_sha256_index ON file_hashes_to_delete (sha256);
            CREATE INDEX IF NOT EXISTS file_hashes_to_delete_deleted_on_index ON file_hashes_to_delete (deleted_on);
            ''')

    def _clean_org_keys(self):
        self._run_sql_query('Removing orgs from deprov list that are not in UBS.',
                            'DELETE FROM orgs_to_deprovision AS otd WHERE otd.deleted_on IS NULL AND NOT EXISTS (SELECT FROM cust WHERE cust.org_key = otd.org_key);')
        self._run_sql_query('Removing orgs from deprov list that are still sending data',
                            '''
        DELETE FROM orgs_to_deprovision 
        WHERE deleted_on IS NULL 
        AND org_key IN (
            SELECT DISTINCT o.org_key
            FROM orgs_to_deprovision o
            JOIN cust c ON o.org_key = c.org_key
            JOIN cust_ix_link cil ON cil.cust_id = c.id
            WHERE cil.date_created > CURRENT_DATE - INTERVAL '1 day');
                            ''')

    @timed_sql_operation
    def _get_org_keys(self):
        logging.info('Getting org_keys to be deprovisioned.')
        self._context.cursor.execute(
            f'''
            SELECT otd.org_key FROM orgs_to_deprovision otd JOIN cust ON cust.org_key = otd.org_key WHERE otd.deleted_on is NULL ORDER BY otd.priority LIMIT {self._context.batch_size};
            '''
        )
        self._org_keys = [record[0] for record in self._context.cursor.fetchall()]
        if not self._org_keys:
            raise(NothingToDoError('There are no orgs to deprovision.'))
        self._context.cursor.executemany('INSERT INTO async_orgs VALUES (%s)', ((key,) for key in self._org_keys))
        return f'The following org_keys were found to deprovision: {self._org_keys} (Count: {len(self._org_keys)})'

    @timed_sql_operation
    def _drop_tables(self):
        logging.info('Dropping tables.')
        self._context.cursor.execute(
            '''
            DROP TABLE IF EXISTS cust_ix_link_to_delete;
            DROP TABLE IF EXISTS device_to_delete;
            DROP TABLE IF EXISTS cust_ids_to_delete;
            DROP TABLE IF EXISTS file_index_to_delete;
            DROP TABLE IF EXISTS cust_file_details_to_delete;
            ''', (self._org_keys,)
        )

    @timed_sql_operation
    def _create_tables(self):
        logging.info('Creating tables.')
        self._context.cursor.execute(
            '''
            CREATE TABLE cust_ix_link_to_delete (LIKE cust_ix_link);
            CREATE TABLE device_to_delete (id BIGINT PRIMARY KEY);
            CREATE TABLE cust_ids_to_delete AS (SELECT id FROM cust WHERE org_key = ANY (%s));
            CREATE TABLE file_index_to_delete (id BIGINT PRIMARY KEY);           
            CREATE TABLE cust_file_details_to_delete(id BIGINT PRIMARY KEY, s3_key VARCHAR(1024) NOT NULL, file_index_id BIGINT, deleted_on TIMESTAMP DEFAULT NULL);
            CREATE INDEX IF NOT EXISTS cust_file_details_to_delete_file_index_id_index ON cust_file_details_to_delete (file_index_id);
            CREATE INDEX IF NOT EXISTS cust_file_details_to_delete_deleted_on_index ON cust_file_details_to_delete (deleted_on);
            ''', (self._org_keys,)
        )

    @timed_sql_operation
    def _find_cust_links_to_delete(self):
        logging.info('Finding all relevant customer links.')
        self._context.cursor.execute(
            '''
            TRUNCATE TABLE cust_ix_link_to_delete;
            INSERT INTO cust_ix_link_to_delete SELECT cil.* FROM cust_ids_to_delete cd JOIN cust_ix_link cil ON cil.cust_id = cd.id;
            '''
        )
        return f'{self._context.cursor.rowcount:,}: Number of customer links.'

    @timed_sql_operation
    def _find_cust_devices_for_table(self, table: str) -> int:
        self._context.cursor.execute('TRUNCATE TABLE device_to_delete;')
        return AsyncBatchJob(AsyncBatchConfig(
                context=self._context,
                name=F'>> Looking in {table}',
                batch_size=config.link_batch_size,
                batched_table='cust_ix_link_to_delete',
                batch_query=F'INSERT INTO device_to_delete SELECT DISTINCT dl.device_id FROM {table} dl JOIN batch_table ON dl.cust_ix_link_id = batch_table.id ON CONFLICT DO NOTHING;''')).run()

    @timed_sql_operation
    def _find_cust_devices(self):
        logging.info('Finding all devices owned by customers.')
        count = self._find_cust_devices_for_table('cust_device_link')
        count += self._find_cust_devices_for_table('cust_device_link_1')
        count += self._find_cust_devices_for_table('cust_device_link_3')
        return f'{count:,}: Number of devices to delete.'

    def _find_unique_files_to_delete(self):
        self._context.cursor.execute('TRUNCATE TABLE file_index_to_delete;')
        self._run_sql_query(
            'Finding unique files to be deleted.',
            '''
            INSERT INTO file_index_to_delete
                SELECT unique_r.file_index_id FROM (
                    SELECT cil.file_index_id, count(*) as count FROM cust_ix_link cil JOIN cust_ix_link_to_delete ciltd on cil.file_index_id = ciltd.file_index_id GROUP BY cil.file_index_id
                ) as unique_r
                WHERE unique_r.count = 1''')

    def _find_s3_files_to_delete(self):
        self._context.cursor.execute('TRUNCATE TABLE cust_file_details_to_delete;')
        return AsyncBatchJob(AsyncBatchConfig(
            context=self._context,
            name='Finding file details to be deleted.',
            batch_size=config.link_batch_size,
            batched_table='file_index_to_delete',
            batch_query='''
            INSERT INTO cust_file_details_to_delete (
                SELECT fd.id, fd.s3_key, fd.file_index_id FROM file_details_cust_link fdcl
                   JOIN file_details fd ON fdcl.file_details_id = fd.id
                   JOIN batch_table ON batch_table.id = fd.file_index_id
                   GROUP BY fd.id);''',
            timed=True)).run()

    @timed_sql_operation
    def _run_sql_query(self, info_message: str, query: str, dry_run: bool = False):
        logging.info(info_message)
        if dry_run:
            self._context.cursor.execute('BEGIN;')

        self._context.cursor.execute(query)
        if self._context.cursor.rowcount is not None:
            return self._context.cursor.rowcount
        one_value = self._context.cursor.fetchone()

        if dry_run:
            self._context.cursor.execute('ROLLBACK;')

        if one_value is not None:
            if one_value[0] < 0:
                return
            return one_value[0]

    def _delete_devices(self):
        self._run_sql_query('Deleting devices.',
                            'DELETE FROM device USING device_to_delete dtd WHERE device.id = dtd.id;',
                            config.dry_run)

    def _delete_file_details(self):
        self._run_sql_query('Deleting file details customer links.',
                            'DELETE FROM file_details_cust_link AS fdcl USING cust_ids_to_delete cd WHERE fdcl.cust_id = cd.id;',
                            config.dry_run)
        return AsyncBatchJob(AsyncBatchConfig(
            context=self._context,
            name='Deleting file details',
            batch_size=config.link_batch_size,
            batched_table='cust_file_details_to_delete',
            batch_query='DELETE FROM file_details USING batch_table WHERE file_details.id = batch_table.id',
            dry_run=config.dry_run,
            timed=True)).run()

    @timed_sql_operation
    def _delete_cust_links(self):
        return AsyncBatchJob(AsyncBatchConfig(
            context=self._context,
            name='Delete customer file index links',
            batch_size=config.link_batch_size,
            pool_size=config.pool_size*2,
            batched_table='cust_ix_link_to_delete',
            batch_query='DELETE FROM cust_ix_link USING batch_table WHERE cust_ix_link.id = batch_table.id',
            dry_run=config.dry_run)).run()

    def _delete_metadata(self):
        AsyncBatchJob(AsyncBatchConfig(
            context=self._context,
            name='Delete metadata',
            batch_size=config.link_batch_size,
            pool_size=config.pool_size*2,
            batched_table='cust_file_details_to_delete',
            batch_query='DELETE FROM metadata USING batch_table WHERE metadata.file_index_id = batch_table.file_index_id',
            dry_run=config.dry_run,
            timed=True)).run()
        AsyncBatchJob(AsyncBatchConfig(
            context=self._context,
            name='Delete file indexes',
            batch_size=config.link_batch_size,
            pool_size=config.pool_size*2,
            batched_table='cust_file_details_to_delete',
            batch_query='DELETE FROM file_index USING batch_table WHERE file_index.id = batch_table.file_index_id',
            dry_run=config.dry_run,
            timed=True)).run()

    def _delete_customers(self):
        self._run_sql_query('Deleting customers.',
                            'DELETE FROM cust USING cust_ids_to_delete citd WHERE cust.id = citd.id;',
                            dry_run=config.dry_run)

    def _copy_file_details_to_delete(self):
        self._run_sql_query('Copying file details to delete for later deletion.',
                            'INSERT INTO file_details_to_delete (SELECT cfdtd.* FROM cust_file_details_to_delete AS cfdtd) ON CONFLICT DO NOTHING;',
                            dry_run=config.dry_run)

    def _copy_file_hashes_to_delete(self):
        self._run_sql_query(
            'Copying file hashes to delete for later deletion.',
            '''
                INSERT INTO file_hashes_to_delete(id, sha256)
                    SELECT fid.id, fi.sha256 FROM file_index_to_delete fid JOIN file_index fi ON fi.id = fid.id ON CONFLICT DO NOTHING;''',
            dry_run=config.dry_run)

    @timed_sql_operation
    def _remove_org_keys_from_deprov_list(self):
        logging.info('Marking deleted org keys in deprov list as deleted.')
        if not config.dry_run:
            self._context.cursor.execute(
                '''
                UPDATE orgs_to_deprovision SET deleted_on = now() WHERE org_key = ANY(%s);
                ''', (self._org_keys,)
            )


def run():
    try:
        with Services(auto_commit=True) as services:
            logging.info('Starting DB execution.')
            Deprovision(services.db_connection, config, DeprovisionRunner).run()
            logging.info('DB execution completed successfully.')
    except NothingToDoError as e:
        logging.info(f'{e}  Exiting...')


def main():
    setup_logging(config)
    config.log()
    if config.continuous_mode:
        logging.info(f'Running in continuous mode restarting after {config.rerun_delay_min} minutes once finished.')
        while True:
            run()
            logging.info(f'Starting next run @ {datetime.datetime.now() + datetime.timedelta(minutes=config.rerun_delay_min)}.')
            sleep(config.rerun_delay_min * 60)
    else:
        run()


if __name__ == "__main__":
    try:
        main()
    except Exception as ex:
        logging.exception(f'Runtime Error ({type(ex)}): {ex}')
        exit(1)
