from abc import ABC, abstractmethod

import sys
from os import environ
import logging

import json_log_formatter

from service import Services

from utils import timed_sql_operation


class Config(object):
    def __init__(self):
        self.batch_size = self.get_int_environ("BATCH_SIZE", 10)
        self.link_batch_size = self.get_int_environ("LINK_BATCH_SIZE", 100000)
        self.dry_run = self.get_bool_environ("DRY_RUN", True)
        self.log_level = self.get_environ("LOG_LEVEL", "INFO")
        self.verbose = self.get_bool_environ("VERBOSE", True)
        self.aws_logging = self.get_bool_environ("AWS_LOGGER", True)

    @staticmethod
    def get_environ(label: str, default: str) -> str:
        return environ.get(label) or default

    @staticmethod
    def get_int_environ(label: str, default: int) -> int:
        return int(environ.get(label) or default)

    @staticmethod
    def get_bool_environ(label: str, default: bool) -> bool:
        return str(environ.get(label) or default).lower() == 'true'

    def log(self):
        logging.info(
            F'''CONFIG: Batch Size: {self.batch_size}, Link Batch Size: {self.link_batch_size}, Dry Run: {self.dry_run}, Log Level: {self.log_level}'''
        )


config = Config()


class DurationFormatter(logging.Formatter):
    def __init__(self):
        super().__init__()
        self._default = logging.Formatter('%(asctime)s - %(levelname)s: %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')
        self._extra = logging.Formatter('%(asctime)s - %(levelname)s [%(duration)s]: %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p')

    def format(self, record: logging.LogRecord) -> str:
        if getattr(record, 'duration', None):
            return self._extra.format(record)
        return self._default.format(record)


def setup_logging():
    formatter = json_log_formatter.JSONFormatter() if config.aws_logging else DurationFormatter()
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)
    handler.setLevel(config.log_level)
    logging.basicConfig(level=config.log_level, handlers=[handler])


class BatchedSqlContext(ABC):
    @property
    @abstractmethod
    def cursor(self):
        pass

    @property
    @abstractmethod
    def batch_size(self) -> int:
        pass

    @property
    @abstractmethod
    def link_batch_size(self) -> int:
        pass

    @property
    @abstractmethod
    def dry_run(self) -> bool:
        pass


class Deprovision(BatchedSqlContext):
    def __init__(self, db_connection, config: 'Config'):
        self._db_connection = db_connection
        self._config = config
        self._cursor = None

    def run(self):
        with self._db_connection.cursor() as cursor:
            self._cursor = cursor
            _DeprovisionRunner(self).run()

    @property
    def cursor(self):
        return self._cursor

    @property
    def batch_size(self) -> int:
        return self._config.batch_size

    @property
    def dry_run(self) -> bool:
        return self._config.dry_run

    @property
    def link_batch_size(self) -> int:
        return self._config.link_batch_size


class _DeprovisionRunner(object):
    def __init__(self, batched_context: BatchedSqlContext):
        self._context = batched_context
        self._org_keys = None

    def run(self):
        if self._context.dry_run:
            logging.info('Performing dry run (will not be committed).')

        self._get_org_keys()
        if not self._org_keys:
            logging.warning('!! No org_keys to deprovision were found.  Exiting.')
            return

        self._create_temp_tables()
        self._find_cust_links_to_delete()
        self._find_cust_devices()
        self._find_unique_files_to_delete()
        self._find_s3_files_to_delete()
        # self._find_unique_signatures()
        # self._find_filepaths()
        # self._find_company_names()
        # self._find_product_names()
        # self._find_copyrights()
        self._delete_devices()
        # self._delete_signatures()
        # self._delete_filepaths()
        self._delete_file_details()
        self._delete_cust_links()
        self._delete_metadata()
        self._delete_customers()
        self._copy_file_details_to_delete()
        self._remove_org_keys_from_deprov_list()
        self._close_transaction()

    @timed_sql_operation
    def _run_batch_job(self, name: str, batch_size: int, count_query: str, temp_query: str, batch_query: str, percent_change: int = 2):
        logging.info(F'Batching: {name}')
        self._context.cursor.execute(count_query)
        end = self._context.cursor.fetchone()[0]
        row_count = 0
        if not end:
            if config.verbose: logging.info(F'{name}: Nothing to loop on.')
            return

        last_percent = 0
        self._context.cursor.execute('CREATE TEMP TABLE batch_table (id BIGINT PRIMARY KEY);')
        for index in range(0, end + 1, batch_size):
            next_percent = index * 100.0 / end
            next_percent = next_percent - next_percent % percent_change
            if next_percent >= last_percent + percent_change:
                last_percent = next_percent
                if config.verbose: logging.info(F'{name}: {last_percent}%')

            self._context.cursor.execute(
                F'''
                INSERT INTO batch_table {temp_query} OFFSET {index} LIMIT {batch_size};
                {batch_query};
                '''
            )
            row_count += self._context.cursor.rowcount
            self._context.cursor.execute('TRUNCATE batch_table;')
        if last_percent < 100:
            if config.verbose: logging.info(F'{name}: 100%')
        self._context.cursor.execute('DROP TABLE batch_table;')
        return row_count

    @timed_sql_operation
    def _get_org_keys(self):
        logging.info('Getting org_keys to be deprovisioned.')
        self._context.cursor.execute(
            f'''
            SELECT otd.org_key FROM orgs_to_deprovision otd JOIN cust ON cust.org_key = otd.org_key WHERE otd.deleted_on is NULL ORDER BY otd.priority LIMIT {self._context.batch_size};
            '''
        )
        self._org_keys = [record[0] for record in self._context.cursor.fetchall()]
        return f'The following org_keys were found to deprovision: {self._org_keys} (Count: {len(self._org_keys)})'

    @timed_sql_operation
    def _create_temp_tables(self):
        logging.info('Creating temporary tables.')
        self._context.cursor.execute(
            '''
            CREATE TEMP TABLE cust_ix_link_to_delete (LIKE cust_ix_link);
            CREATE TEMP TABLE device_to_delete (id BIGINT PRIMARY KEY);
            CREATE TEMP TABLE cust_ids_to_delete AS (SELECT id FROM cust WHERE org_key = ANY (%s));
            CREATE TEMP TABLE file_index_to_delete (id BIGINT PRIMARY KEY);
            CREATE TEMP TABLE signature_to_delete (id BIGINT PRIMARY KEY);
            CREATE TEMP TABLE filepath_to_delete (id BIGINT PRIMARY KEY);
            CREATE TEMP TABLE company_name_to_delete (id BIGINT PRIMARY KEY);
            CREATE TEMP TABLE program_name_to_delete (id BIGINT PRIMARY KEY);
            CREATE TEMP TABLE copyright_to_delete (id BIGINT PRIMARY KEY);
            CREATE TEMP TABLE org_filepaths (id BIGINT PRIMARY KEY);
            CREATE TEMP TABLE unique_filepaths (id BIGINT PRIMARY KEY);
            CREATE TEMP TABLE cust_file_details_to_delete(id BIGINT PRIMARY KEY, s3_key VARCHAR(1024) NOT NULL, file_index_id BIGINT, deleted_on TIMESTAMP DEFAULT NULL);
            CREATE INDEX IF NOT EXISTS cust_file_details_to_delete_file_index_id_index ON cust_file_details_to_delete (file_index_id);
            CREATE INDEX IF NOT EXISTS cust_file_details_to_delete_deleted_on_index ON cust_file_details_to_delete (deleted_on);
            ''', (self._org_keys,)
        )

    @timed_sql_operation
    def _find_cust_links_to_delete(self):
        logging.info('Finding all relevant customer links.')
        self._context.cursor.execute(
            '''
            INSERT INTO cust_ix_link_to_delete SELECT cil.* FROM cust_ids_to_delete cd JOIN cust_ix_link cil ON cil.cust_id = cd.id;
            '''
        )
        return f'{self._context.cursor.rowcount}: Number of customer links.'

    # def _find_cust_devices_for_table(self, table: str, previous_count: int) -> int:
    #     return self._run_batch_job(
    #         name=F'>> Looking in {table}',
    #         batch_size=10000,
    #         count_query='SELECT count(*) FROM cust_ix_link_to_delete',
    #         temp_query='SELECT id FROM cust_ix_link_to_delete',
    #         batch_query=F'INSERT INTO device_to_delete SELECT DISTINCT dl.device_id FROM {table} dl JOIN batch_table ON dl.cust_ix_link_id = batch_table.id ON CONFLICT DO NOTHING')

    @timed_sql_operation
    def _find_cust_devices_for_table(self, table: str, previous_count: int) -> int:
        logging.info(F'>> Looking in {table}')
        self._context.cursor.execute(
            F'''
            CALL get_devices_to_delete('{table}', {self._context.link_batch_size});
            SELECT COUNT(*) FROM device_to_delete;
            '''
        )
        return self._context.cursor.fetchone()[0] - previous_count

    @timed_sql_operation
    def _find_cust_devices(self):
        logging.info('Finding all devices owned by customers.')
        count = self._find_cust_devices_for_table('cust_device_link', 0)
        count += self._find_cust_devices_for_table('cust_device_link_1', count)
        count += self._find_cust_devices_for_table('cust_device_link_3', count)
        return f'{count}: Number of devices to delete.'

    @timed_sql_operation
    def _find_unique_files_to_delete(self):
        logging.info('Finding unique files to be deleted.')
        self._context.cursor.execute(
            '''
            INSERT INTO file_index_to_delete
                SELECT unique_r.file_index_id FROM (
                    SELECT file_index_id, count(*) as count FROM cust_ix_link cil GROUP BY cil.file_index_id
                ) as unique_r
                JOIN cust_ix_link_to_delete cil ON cil.file_index_id = unique_r.file_index_id
                WHERE unique_r.count = 1;
            '''
        )
        return f'{self._context.cursor.rowcount}: Number of files to delete.'

    @timed_sql_operation
    def _find_s3_files_to_delete(self):
        logging.info('Finding s3 links to files to be deleted.')
        self._context.cursor.execute(
            '''
            INSERT INTO cust_file_details_to_delete (
                SELECT fd.id, fd.s3_key, fd.file_index_id FROM file_details_cust_link fdcl
                   JOIN file_details fd ON fdcl.file_details_id = fd.id
                   JOIN file_index_to_delete ui ON ui.id = fd.file_index_id
                GROUP BY fd.id
            );
            '''
        )
        return f'{self._context.cursor.rowcount}: Number of s3 links to delete.'

    @timed_sql_operation
    def _find_unique_signatures(self):
        logging.info('Finding unique signatures to be deleted.')
        self._context.cursor.execute(
            '''
            INSERT INTO signature_to_delete
                SELECT unique_r.signature_id FROM (
                    -- All the unique signatures.
                    SELECT signature_id, count(*) as count FROM signature_result_link srl GROUP BY srl.signature_id
                ) as unique_r
                -- Link to customer index link table
                JOIN signature_result_link srl ON srl.signature_id = unique_r.signature_id
                -- Link to links to be deleted so we only delete what belongs to these customers.
                JOIN cust_ix_link_to_delete cil on srl.cust_ix_link_id = cil.id
                -- For only results which are actually unique.
                WHERE unique_r.count = 1;
            '''
        )
        return f'{self._context.cursor.rowcount}: Number of signatures to delete.'

    @timed_sql_operation
    def _fill_org_filepaths(self):
        logging.info('Finding all customer filepaths.')
        self._context.cursor.execute(
            '''
            CALL get_filepaths(10000);
            SELECT count(*) FROM org_filepaths;
            '''
        )
        return self._context.cursor.fetchone()[0]

    @timed_sql_operation
    def _find_unique_filepaths(self):
        logging.info('Finding unique filepaths to be deleted.')
        self._context.cursor.execute(
            '''
            INSERT INTO filepath_to_delete
                SELECT unique_r.filepath_id FROM (
                    -- All the unique filepaths.
                    SELECT filepath_id, count(*) as count FROM cust_filepath cf JOIN org_filepaths ON org_filepaths.id = cf.filepath_id GROUP BY cf.filepath_id
                ) as unique_r
                WHERE unique_r.count = 1;
            '''
        )
        return f'{self._context.cursor.rowcount}: Number of filepaths to delete.'

    def _find_filepaths(self):
        self._fill_org_filepaths()
        self._find_unique_filepaths()

    @timed_sql_operation
    def _find_company_names(self):
        logging.info('Finding company_name to be deleted.')
        self._context.cursor.execute(
            '''
            INSERT INTO company_name_to_delete
                SELECT unique_r.company_name_id FROM (
                     -- All the unique company names.
                     SELECT company_name_id, count(*) as count FROM metadata m GROUP BY company_name_id
                ) as unique_r
                JOIN metadata m ON m.company_name_id = unique_r.company_name_id
                JOIN file_index_to_delete fdtd ON fdtd.id = m.file_index_id
                WHERE unique_r.count = 1;
            '''
        )
        return f'{self._context.cursor.rowcount}: Number of company_name to delete.'

    @timed_sql_operation
    def _find_product_names(self):
        logging.info('Finding product_name & internal_name to be deleted.')
        self._context.cursor.execute(
            '''
            INSERT INTO program_name_to_delete
                SELECT unique_r.product_name_id FROM (
                   SELECT product_name_id, count(*) as count FROM metadata m GROUP BY product_name_id
                ) as unique_r
                JOIN metadata m ON m.product_name_id = unique_r.product_name_id
                JOIN file_index_to_delete fdtd ON fdtd.id = m.file_index_id
                WHERE unique_r.count = 1;
            '''
        )
        rowcount = self._context.cursor.rowcount
        self._context.cursor.execute(
            '''
            INSERT INTO program_name_to_delete
                SELECT unique_r.internal_name_id FROM (
                   SELECT internal_name_id, count(*) as count FROM metadata m GROUP BY internal_name_id
                ) as unique_r
                JOIN metadata m ON m.internal_name_id = unique_r.internal_name_id
                JOIN file_index_to_delete fdtd ON fdtd.id = m.file_index_id
                WHERE unique_r.count = 1;
            '''
        )
        return f'{self._context.cursor.rowcount + rowcount}: Number of product_name & internal_name to delete.'

    @timed_sql_operation
    def _find_copyrights(self):
        logging.info('Finding copyright & trademark to be deleted.')
        self._context.cursor.execute(
            '''
            INSERT INTO copyright_to_delete
                SELECT unique_r.copyright_id FROM (
                   SELECT copyright_id, count(*) as count FROM metadata m GROUP BY copyright_id
                ) as unique_r
                JOIN metadata m ON m.copyright_id = unique_r.copyright_id
                JOIN file_index_to_delete fdtd ON fdtd.id = m.file_index_id
                WHERE unique_r.count = 1;
            '''
        )
        rowcount = self._context.cursor.rowcount
        self._context.cursor.execute(
            '''
            INSERT INTO copyright_to_delete
                SELECT unique_r.trademark_id FROM (
                    SELECT trademark_id, count(*) as count FROM metadata m GROUP BY trademark_id
                ) as unique_r
                JOIN metadata m ON m.trademark_id = unique_r.trademark_id
                JOIN file_index_to_delete fdtd ON fdtd.id = m.file_index_id
                WHERE unique_r.count = 1;
            '''
        )
        return f'{self._context.cursor.rowcount + rowcount}: Number of copyright & trademark to delete.'

    @timed_sql_operation
    def _run_sql_query(self, info_message: str, query: str):
        logging.info(info_message)
        self._context.cursor.execute(query)
        if self._context.cursor.rowcount is not None:
            return self._context.cursor.rowcount
        one_value = self._context.cursor.fetchone()
        if one_value is not None:
            return one_value[0]

    def _delete_devices(self):
        self._run_sql_query('Deleting devices.',
                            'DELETE FROM device USING device_to_delete dtd WHERE device.id = dtd.id;')

    # def _delete_signatures(self):
    #     self._run_batch_job(
    #         name='Delete signature result links',
    #         batch_size=10000,
    #         count_query='SELECT count(*) FROM cust_ix_link_to_delete',
    #         temp_query='SELECT id FROM cust_ix_link_to_delete',
    #         batch_query='DELETE FROM signature_result_link AS srl USING batch_table WHERE srl.cust_ix_link_id = batch_table.id')
    #     # self._run_sql_query('Deleting signatures.',
    #     #                     'DELETE FROM signature AS sig USING signature_to_delete us WHERE sig.id = us.id;')

    def _delete_signatures(self):
        self._run_sql_query('Deleting signature result links.',
                            'DELETE FROM signature_result_link AS srl USING cust_ix_link_to_delete cil WHERE srl.cust_ix_link_id = cil.id;')
        # self._run_sql_query('Deleting signatures.',
        #                     'DELETE FROM signature AS sig USING signature_to_delete us WHERE sig.id = us.id;')

    def _delete_filepaths(self):
        self._run_batch_job(
            name='Delete customer filepaths',
            batch_size=10000,
            count_query='SELECT count(*) FROM cust_ix_link_to_delete',
            temp_query='SELECT id FROM cust_ix_link_to_delete',
            batch_query='DELETE FROM cust_filepath AS cf USING batch_table WHERE cf.cust_ix_link_id = batch_table.id')
        # self._run_sql_query('Deleting filepaths.',
        #                     'DELETE FROM filepath USING filepath_to_delete fdtd WHERE filepath.id = fdtd.id;')

    # def _delete_filepaths(self):
    #     # self._run_sql_query('Getting count of customer filepath links.',
    #     #                     'SELECT count(*) FROM cust_filepath AS cf JOIN cust_ix_link_to_delete cil ON cf.cust_ix_link_id = cil.id;')
    #     self._run_sql_query('Deleting customer filepaths.',
    #                         'CALL delete_cust_filepaths(10000);')
    #     # self._run_sql_query('Deleting filepaths.',
    #     #                     'DELETE FROM filepath USING filepath_to_delete fdtd WHERE filepath.id = fdtd.id;')

    def _delete_file_details(self):
        # self._run_sql_query('Deleting file details magic links.',
        #                     'DELETE FROM file_details_magic_link AS fdml USING file_details_to_delete fdtd WHERE fdml.file_details_id = fdtd.id;')
        # self._run_sql_query('Deleting file details partner links.',
        #                     'DELETE FROM file_details_partner_link AS fdpl USING file_details_to_delete fdtd WHERE fdpl.file_details_id = fdtd.id;')
        self._run_sql_query('Deleting file details customer links.',
                            'DELETE FROM file_details_cust_link AS fdcl USING cust_ids_to_delete cd WHERE fdcl.cust_id = cd.id;')
        self._run_sql_query('Deleting file details.',
                            'DELETE FROM file_details AS fd USING cust_file_details_to_delete cd WHERE fd.id = cd.id;')

    def _delete_cust_links(self):
        # self._run_batch_job(
        #     name='Delete customer file index links',
        #     batch_size=10000,
        #     count_query='SELECT count(*) FROM cust_ix_link_to_delete',
        #     temp_query='SELECT id FROM cust_ix_link_to_delete',
        #     batch_query='DELETE FROM cust_ix_link AS cf USING batch_table WHERE cf.id = batch_table.id')
        self._run_sql_query('Deleting customer file index links.',
                            'DELETE FROM cust_ix_link USING cust_ids_to_delete cd WHERE cust_ix_link.cust_id = cd.id;')

    def _delete_metadata(self):
        self._run_sql_query('Deleting metadata.',
                            'DELETE FROM metadata AS md USING cust_file_details_to_delete fdtd WHERE md.file_index_id = fdtd.file_index_id;')
        self._run_sql_query('Deleting file indexes.',
                            'DELETE FROM file_index AS fi USING cust_file_details_to_delete fdtd WHERE fdtd.file_index_id = fi.id;')
        # self._run_sql_query('Deleting company names.',
        #                     'DELETE FROM company_name AS cn USING company_name_to_delete cntd WHERE cn.id = cntd.id;')
        # self._run_sql_query('Deleting program names.',
        #                     'DELETE FROM program_name AS pn USING program_name_to_delete pntd WHERE pn.id = pntd.id;')
        # self._run_sql_query('Deleting copyrights.',
        #                     'DELETE FROM copyright AS cn USING copyright_to_delete cntd WHERE cn.id = cntd.id;')

    def _delete_customers(self):
        self._run_sql_query('Deleting customers.',
                            'DELETE FROM cust USING cust_ids_to_delete citd WHERE cust.id = citd.id;')

    def _copy_file_details_to_delete(self):
        self._run_sql_query('Copying file details to delete for later deletion.',
                            'INSERT INTO file_details_to_delete (SELECT cfdtd.* FROM cust_file_details_to_delete AS cfdtd) ON CONFLICT DO NOTHING;')

    @timed_sql_operation
    def _remove_org_keys_from_deprov_list(self):
        logging.info('Marking deleted org keys in deprov list as deleted.')
        self._context.cursor.execute(
            '''
            UPDATE orgs_to_deprovision SET deleted_on = now() WHERE org_key = ANY(%s);
            ''', (self._org_keys,)
        )

    @timed_sql_operation
    def _close_transaction(self):
        if self._context.dry_run:
            logging.info('This was a dry run, rolling back changes.')
            self._context.cursor.execute('ROLLBACK;')
        else:
            logging.info('Committing changes.')
            self._context.cursor.execute('COMMIT;')


def main():
    setup_logging()
    config.log()

    with Services() as services:
        logging.info('Starting DB execution.')
        Deprovision(services.db_connection, config).run()
        logging.info('DB execution completed successfully.')


if __name__ == "__main__":
    try:
        main()
    except Exception as ex:
        logging.exception(f'Runtime Error ({type(ex)}): {ex}')
        exit(1)
