import datetime
import logging
from os import environ

import boto3
import psycopg2

from aws.kms_client import KmsClient
from aws.s3_client import S3Client


def to_bool(param):
    return str(param).lower() in ("true", "1")


def decrypt_password(password: str, decrypt: bool, aws_region: str) -> str:
    if decrypt:
        kms_client = KmsClient(boto3.Session(region_name=aws_region))
        password = kms_client.decrypt_string(password)

    return password


def connect_to_db(aws_region: str):
    return psycopg2.connect(
        host=environ["ubs_db_writer_host"],
        port=int(environ["ubs_db_port"]),
        database=environ["ubs_db_db_name"],
        user=environ["ubs_db_user"],
        password=decrypt_password(environ["ubs_db_pw"], to_bool(environ["ubs_db_pw_encrypted"]), aws_region)
    )


def get_fips_3s_endpoint() -> str:
    return "https://" + environ["AWS_ENDPOINT_OVERRIDE_S3"] if "AWS_ENDPOINT_OVERRIDE_S3" in environ else None


class Services(object):
    def __init__(self, auto_commit: bool = False):
        self._kms_client = None
        self._s3_client = None
        self._aws_region = environ.get("AWS_DEFAULT_REGION", "us-east-1")
        self._db_connection = None
        self._start_time = None
        self._auto_commit = auto_commit
        self._opened = False

    def open(self) -> 'Services':
        logging.info('Connecting to db.')
        if self._opened:
            raise RuntimeError('Connection to the database is already open.')
        self._start_time = datetime.datetime.now()
        self._db_connection = connect_to_db(self._aws_region)
        self._db_connection.autocommit = self._auto_commit
        self._opened = True
        return self

    def close(self):
        logging.info('Closing db connection.')
        if self._opened:
            try:
                self._db_connection.close()
                self._db_connection = None
            finally:
                logging.info(F'Total time session was open: {datetime.datetime.now() - self._start_time}')
                self._opened = False

    def s3_client(self) -> S3Client:
        return S3Client(boto_session=boto3.Session(region_name=self._aws_region),
                        endpoint_url=get_fips_3s_endpoint())

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def db_connection(self):
        return self._db_connection
