import logging
from abc import ABC, abstractmethod
from typing import List, Callable


class StageStorageManager(ABC):
    @abstractmethod
    def load(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def save(self, value: int):
        raise NotImplementedError()

    @abstractmethod
    def clear(self):
        raise NotImplementedError()


class DbStageStorageManager(StageStorageManager):
    def __init__(self, db_cursor, table_name='staged_run'):
        self._db_cursor = db_cursor
        self._table_name = table_name

    def load(self) -> int:
        self._db_cursor.execute(f'CREATE TABLE IF NOT EXISTS {self._table_name}(id INT PRIMARY KEY, stage INT);')
        self._db_cursor.execute(f'SELECT stage FROM {self._table_name};')
        result = self._db_cursor.fetchone()
        return result[0] if result else 0

    def save(self, value: int):
        self._db_cursor.execute(f'INSERT INTO {self._table_name} (id, stage) VALUES (0, {value}) ON CONFLICT (id) DO UPDATE SET stage = {value};')

    def clear(self):
        self.save(0)


class StageFunction(object):
    def __init__(self, func: Callable, args, kwargs):
        self._func = func
        self._args = args
        self._kwargs = kwargs

    def run(self):
        self._func(*self._args, **self._kwargs)


class Stage(object):
    def __init__(self, index: int, label: str):
        self._index = index
        self._label = label
        self._functions: List[StageFunction] = []

    def append(self, func: Callable, *args, **kwargs):
        self._functions.append(StageFunction(func, args, kwargs))

    @property
    def index(self) -> int:
        return self._index

    @property
    def label(self) -> str:
        return self._label

    def run(self):
        for func in self._functions:
            func.run()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __str__(self):
        return f'Stage {self._index}: {self._label}'


class StagedRun(object):
    def __init__(self, stage_storage: StageStorageManager):
        self._storage = stage_storage
        self._stages: List[Stage] = []

    def add_stage(self, label: str) -> Stage:
        new_stage = Stage(len(self._stages), label)
        self._stages.append(new_stage)
        return new_stage

    def _load_stage_index(self) -> int:
        stage_index = self._storage.load()
        if not stage_index:
            logging.info('Performing fresh run.')
        else:
            logging.info(f'Continuing run from {self._stages[stage_index]}.')
        return stage_index

    def run(self):
        start_index = self._load_stage_index()
        for stage in self._stages[:start_index]:
            logging.info(f'Skipping {stage}.')
        for stage in self._stages[start_index:]:
            self._storage.save(stage.index)
            stage.run()
        self._storage.clear()

    def in_progress(self) -> bool:
        return bool(self._storage.load())
